import os
import re
import time
import unittest
from datetime import datetime

from pybrowscap import Browser
from django_pybrowscap import middleware


class RequestMock(object):

    def __init__(self):
        self.path_info = '/'
        self.META = {'HTTP_USER_AGENT': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:10.0.2) Gecko/20100101 Firefox/10.0.2'}


class BaseTest(unittest.TestCase):

    def setUp(self):

        middleware.settings.PYBROWSCAP_FILE_PATH = os.path.join(os.path.dirname(__file__), 'data', 'browscap_07_10_2011.csv')
        self.request = RequestMock()
        self.DEFAULT_PYBROWSCAP_INITIALIZE = middleware.settings.PYBROWSCAP_INITIALIZE
        self.DEFAULT_PYBROWSCAP_FILE_PATH = middleware.settings.PYBROWSCAP_FILE_PATH
        self.DEFAULT_PYBROWSCAP_UPDATE = middleware.settings.PYBROWSCAP_UPDATE
        self.DEFAULT_PYBROWSCAP_UPDATE_INTERVAL = middleware.settings.PYBROWSCAP_UPDATE_INTERVAL
        self.DEFAULT_PYBROWSCAP_IGNORE_PATHS = middleware.settings.PYBROWSCAP_IGNORE_PATHS

    def tearDown(self):
        middleware.settings.PYBROWSCAP_INITIALIZE = self.DEFAULT_PYBROWSCAP_INITIALIZE
        middleware.settings.PYBROWSCAP_FILE_PATH = self.DEFAULT_PYBROWSCAP_FILE_PATH
        middleware.settings.PYBROWSCAP_UPDATE = self.DEFAULT_PYBROWSCAP_UPDATE
        middleware.settings.PYBROWSCAP_UPDATE_INTERVAL = self.DEFAULT_PYBROWSCAP_UPDATE_INTERVAL
        middleware.settings.PYBROWSCAP_IGNORE_PATHS = self.DEFAULT_PYBROWSCAP_IGNORE_PATHS


class TestInitialize(BaseTest):

    def test_initialization_on(self):
        middleware.settings.PYBROWSCAP_INITIALIZE = True
        instance = middleware.PybrowscapMiddleware()
        self.assertTrue(hasattr(instance, 'browscap'))
        instance.process_request(self.request)
        self.assertIsNotNone(self.request.browser)
        self.assertIsInstance(self.request.browser, Browser)
        self.assertFalse(self.request.browser.is_crawler())

    def test_initialization_off(self):
        middleware.settings.PYBROWSCAP_INITIALIZE = False
        instance = middleware.PybrowscapMiddleware()
        self.assertFalse(hasattr(instance, 'browscap'))
        self.assertFalse(hasattr(self.request, 'browser'))

    def test_initialization_no_file(self):
        middleware.settings.PYBROWSCAP_FILE_PATH = ''
        instance = middleware.PybrowscapMiddleware()
        self.assertIsNone(instance.browscap)


class TestUpdate(BaseTest):

    def test_update(self):
        middleware.settings.PYBROWSCAP_UPDATE = True
        middleware.settings.PYBROWSCAP_UPDATE_INTERVAL = 0.1
        instance = middleware.PybrowscapMiddleware()
        self.assertGreaterEqual(datetime.now(), instance.browscap.loaded_at)
        self.assertIsNone(instance.browscap.reloaded_at)
        time.sleep(0.2)
        instance.process_request(self.request)
        self.assertGreaterEqual(datetime.now(), instance.browscap.loaded_at)
        self.assertGreaterEqual(datetime.now(), instance.browscap.reloaded_at)

    def test_update_long_interval(self):
        middleware.settings.PYBROWSCAP_UPDATE = True
        middleware.settings.PYBROWSCAP_UPDATE_INTERVAL = 100
        instance = middleware.PybrowscapMiddleware()
        self.assertGreaterEqual(datetime.now(), instance.browscap.loaded_at)
        self.assertIsNone(instance.browscap.reloaded_at)
        time.sleep(0.1)
        instance.process_request(self.request)
        self.assertGreaterEqual(datetime.now(), instance.browscap.loaded_at)
        self.assertIsNone(instance.browscap.reloaded_at)

    def test_no_update(self):
        middleware.settings.PYBROWSCAP_UPDATE = False
        middleware.settings.PYBROWSCAP_UPDATE_INTERVAL = 0.1
        instance = middleware.PybrowscapMiddleware()
        self.assertGreaterEqual(datetime.now(), instance.browscap.loaded_at)
        self.assertIsNone(instance.browscap.reloaded_at)
        time.sleep(0.1)
        instance.process_request(self.request)
        self.assertGreaterEqual(datetime.now(), instance.browscap.loaded_at)
        self.assertIsNone(instance.browscap.reloaded_at)


class TestIgnorePaths(BaseTest):

    def test_ignore_nothing(self):
        instance = middleware.PybrowscapMiddleware()
        instance.process_request(self.request)
        self.assertIsNotNone(self.request.browser)
        self.assertIsInstance(self.request.browser, Browser)
        self.assertFalse(self.request.browser.is_crawler())

    def test_ignore_media(self):
        middleware.settings.PYBROWSCAP_IGNORE_PATHS = (
            re.compile(r'^/media/'),
        )
        self.request.path_info = '/media/image.jpg'
        instance = middleware.PybrowscapMiddleware()
        instance.process_request(self.request)
        self.assertFalse(hasattr(self.request, 'browser'))


class TestRequestErrors(BaseTest):

    def test_no_path_info(self):
        del self.request.path_info
        instance = middleware.PybrowscapMiddleware()
        instance.process_request(self.request)
        self.assertFalse(hasattr(self.request, 'browser'))

    def test_no_meta(self):
        del self.request.META
        instance = middleware.PybrowscapMiddleware()
        instance.process_request(self.request)
        self.assertTrue(hasattr(self.request, 'browser'))
        self.assertIsNone(self.request.browser)

    def test_no_meta_user_agent(self):
        self.request.META = {}
        instance = middleware.PybrowscapMiddleware()
        instance.process_request(self.request)
        self.assertTrue(hasattr(self.request, 'browser'))
        self.assertIsNone(self.request.browser)


import urllib2
import logging
from datetime import datetime

from django_pybrowscap import settings
from pybrowscap.loader.csv import load_file, URL
from pybrowscap.loader import Downloader


log = logging.getLogger(__name__)


class PybrowscapMiddleware(object):
    """Django middleware class for pybrowscap package."""

    def __init__(self):
        """Constructor.

        This constructor will be called only once while django initialization.

        """
        if settings.PYBROWSCAP_INITIALIZE:
            try:
                log.info('Initializing pybrowscap')
                self.browscap = load_file(settings.PYBROWSCAP_FILE_PATH)
            except IOError:
                log.exception('Error while initializing pybrowscap')
                self.browscap = None
            else:
                log.info('Pybrowscap initialized')

    def process_request(self, request):
        if settings.PYBROWSCAP_INITIALIZE:
            try:
                for regex in settings.PYBROWSCAP_IGNORE_PATHS:
                    if regex.search(request.path_info):
                        return
            except AttributeError:
                log.warn('Invalid request, no path info present')
                return

            try:
                if self.browscap is not None:
                    request.browser = self.browscap.search(request.META['HTTP_USER_AGENT'])
            except AttributeError:
                log.warn('Request has no meta info, impossible to search for user agent')
            except KeyError:
                log.warn('Request has no user agent meta info, impossible to search for user agent')
            finally:
                if not hasattr(request, 'browser'):
                    request.browser = None

            if settings.PYBROWSCAP_UPDATE:
                try:
                    last_reloaded = self.browscap.reloaded_at or self.browscap.loaded_at
                    if (datetime.now() - last_reloaded).total_seconds() > settings.PYBROWSCAP_UPDATE_INTERVAL:
                        try:
                            log.info('Reloading pybrowscap with new data')
                            Downloader(URL).get(settings.PYBROWSCAP_FILE_PATH)
                            self.browscap.reload(settings.PYBROWSCAP_FILE_PATH)
                        except (ValueError, urllib2.HTTPError, urllib2.URLError, IOError):
                            log.exception('Error while reloading pybrowscap')
                        else:
                            log.info('Pybrowscap successfully reloaded')
                except AttributeError:
                    log.warn('Error while reloading uninitialized pybrowscap')


from django.conf import settings

# This setting will effectively turn the middleware off, to speed up requests/response while developing
PYBROWSCAP_INITIALIZE = getattr(settings, 'PYBROWSCAP_INITIALIZE', not settings.DEBUG)

# Path where browscap file is located on filesystem
PYBROWSCAP_FILE_PATH = getattr(settings, 'PYBROWSCAP_FILE_PATH', '')

# Whether to perform automatic updates of browscap file
PYBROWSCAP_UPDATE = getattr(settings, 'PYBROWSCAP_UPDATE', False)

# Interval of automatic browscap file updates
PYBROWSCAP_UPDATE_INTERVAL = getattr(settings, 'PYBROWSCAP_UPDATE_INTERVAL', 604800) # default one week on seconds

# Tuple or regex expressions of path that are to be ignored by middleware
PYBROWSCAP_IGNORE_PATHS = getattr(settings, 'PYBROWSCAP_IGNORE_PATHS', ())